require 'class'

-- FIELDS
local Tank = require 'tank'
local Arena = require 'arena'
local _TankCount = 0
local _UI = {}
local _TankList = {}
local _TurnCount = 0
local _RunLoop = true
local _StepMode = false
local BULLET_SPEED = 20
local BULLET_DAMAGE = 1
local _Arena = {}

-- CONSTRUCTOR
local M=class(function(self, playerList, arenaBuffer)
	-- Create the arena
	_Arena = Arena()
	table.insert(_UI, _Arena)
	-- Create the tanks
	local TANK_HIT_POINTS = 10
	local XCoordiantes = permute(getRangeTable(arenaBuffer, love.graphics.getWidth() - arenaBuffer), nil, #playerList)
	local YCoordiantes = permute(getRangeTable(arenaBuffer, love.graphics.getHeight() - arenaBuffer), nil, #playerList)
	for iter = 1, #playerList, 1 do
		_TankCount = _TankCount + 1
		playerBrain = require(playerList[iter])
		local aTank = Tank(XCoordiantes[_TankCount], YCoordiantes[_TankCount], playerBrain(love.graphics.getWidth() - arenaBuffer, love.graphics.getHeight() - arenaBuffer, arenaBuffer), TANK_HIT_POINTS)
		table.insert(_UI, aTank)
		table.insert(_TankList, aTank)
	end
	_BulletDrawable = love.graphics.newImage("Bullet.png")
	math.randomseed(os.time())
end)

-- ACCESSOR
function M:getUI()
	return _UI
end

-- STAGE FRAMEWORK
function M:update(dt)
	-- Should the game keep processing?
	if _RunLoop then
		-- Yes
		-- Is the game in StepMode/debugging?
		if _StepMode then
			_RunLoop = false
		end
		_TurnCount = _TurnCount + 1
		-- Tell each tank to run its brain
		for iter = 1, #_TankList, 1 do
			if _TankList[iter].hitPoints > 0 then
				-- Generate this tanks the view of the arena
				local tankData = {}
				table.insert(tankData, {_TankList[iter].X, _TankList[iter].Y, _TankList[iter].bodyRotation, _TankList[iter].turretRotation, _TankList[iter].hitPoints, _TankList[iter].X + (-5 * math.sin(_TankList[iter].bodyRotation)), _TankList[iter].Y + (5 * math.cos(_TankList[iter].bodyRotation))})
				for tankIter = 1, #_TankList, 1 do
					if iter ~= tankIter then
						local alive = false
						if _TankList[tankIter].hitPoints > 0 then
							alive = true
						end
						local wasHit = false
						if _TankList[tankIter].wasHit then
							wasHit = true
							_TankList[tankIter].wasHit = false
						end
						table.insert(tankData, {alive, _TankList[tankIter].X, _TankList[tankIter].Y, wasHit})
					end
				end
				-- Ask the tank to make its moves
				_TankList[iter].playerCode:process(tankData)
			end
		end
		-- Process the existing bullets (in flight)
		local bulletRemoveList = {}
		for iter = 1, #_BulletList, 1 do
			-- Detect if we collide			
			-- Are there any tanks close to this bullets flight path?
			local closeObjectList = {}
			for tankIter = 1, #_TankList, 1 do
				-- Is this tank alive/running?
				if _TankList[tankIter].hitPoints > 0 then
					-- Yes
					-- Did this tank NOT fire this bullet?
					if _BulletList[iter][4] ~= tankIter then
						-- Yes, this tank did not fire the bullet
						-- Is a tank close to the bullet?
						if isObjectClose(_BulletList[iter][1], _BulletList[iter][2], _TankList[tankIter].X, _TankList[tankIter].Y) then
							-- Yes, tank is close (within bullet "speed")
							table.insert(closeObjectList, tankIter)
						end
					end
				end
			end
			-- Are any objects close?
			if #closeObjectList > 0 then
				-- Yes, objects are close evaluate which, if any, got hit
				local hitList = {}
				-- For each point in the flight of the bullet
				for pointIter = 1, BULLET_SPEED, 1 do
					local bulletX = _BulletList[iter][1]
					local bulletY = _BulletList[iter][2]
					for closeIter = 1, #closeObjectList, 1 do
						bulletX = bulletX + (pointIter * math.sin(_BulletList[iter][3]))
						bulletY = bulletY + (-pointIter * math.cos(_BulletList[iter][3]))
						-- Did the bullet hit the object?
						if didCollideWithObject(bulletX, bulletY, _TankList[closeObjectList[closeIter]].X, _TankList[closeObjectList[closeIter]].Y, _TankList[closeObjectList[closeIter]].size / 2) then
							-- Yes
							-- Add object to confirmed hit list
							insertIfNotPresent(hitList, closeObjectList[closeIter])
						end
					end
				end
				-- Process hits
				-- Did the bullet hit a tank?
				if #hitList > 0 then
					-- Yes
					-- Find the first/closest object the bullet hit
					local closestHitIndex = 0
					local closestHitIndexDistance = 10000
					for hitIter = 1, #hitList, 1 do
						-- Is this tank closer than the last one (therefore hit first)?
						if distanceBetweenPoints(_BulletList[iter][1], _BulletList[iter][2], _TankList[hitList[hitIter]].X, _TankList[hitList[hitIter]].Y) < closestHitIndexDistance then
							-- Yes
							closestHitIndex = hitList[hitIter]
							closestHitIndexDistance = distanceBetweenPoints(_BulletList[iter][1], _BulletList[iter][2], _TankList[hitList[hitIter]].X, _TankList[hitList[hitIter]].Y)
						end
					end
					table.insert(bulletRemoveList, iter)
					_TankList[closestHitIndex].hitPoints = _TankList[closestHitIndex].hitPoints - 1
					_TankList[closestHitIndex].wasHit = true
					_TankList[closestHitIndex].turnKilled = _TurnCount
				else
					-- No, the bullet missed
					-- Move the bullet to the final location
					_BulletList[iter][1] = _BulletList[iter][1] + (BULLET_SPEED * math.sin(_BulletList[iter][3]))
					_BulletList[iter][2] = _BulletList[iter][2] + (-BULLET_SPEED * math.cos(_BulletList[iter][3]))
				end
			else
				-- No, no objects are close
				-- Move the bullet to the final location
				_BulletList[iter][1] = _BulletList[iter][1] + (BULLET_SPEED * math.sin(_BulletList[iter][3]))
				_BulletList[iter][2] = _BulletList[iter][2] + (-BULLET_SPEED * math.cos(_BulletList[iter][3]))
			end
			-- Remove all bullets outside the "play" walls
			if _BulletList[iter][1] < 0 or _BulletList[iter][1] > love.graphics.getWidth() or _BulletList[iter][2] < 0 or _BulletList[iter][2] > love.graphics.getHeight() then
				-- Yes, remove it from the list
				table.insert(bulletRemoveList, iter)
			end
		end
		-- Count the dead tanks
		local deadCount = 0
		for iter = 1, #_TankList, 1 do
			if _TankList[iter].hitPoints <= 0 then
				deadCount = deadCount + 1
			end
		end
		-- Are we down to one or less tanks (is the game over)?
		if deadCount >= (_TankCount - 1) then
			-- Yes
			_RunLoop = false
			for iter = 1, #_TankList, 1 do
				if (_TankList[iter].turnKilled == _TurnCount) and (_TankList[iter].hitPoints <= 0) then
					_TankList[iter].hitPoints = .5
				end
			end
			return
		end
		-- Remove spent bullets
		if #bulletRemoveList > 0 then
			quicksort(bulletRemoveList)
			for iter = #bulletRemoveList, 1, -1 do
				table.remove(_BulletList, bulletRemoveList[iter])
			end
		end
		-- Process the results (of each tanks move)
		for iter = 1, #_TankList, 1 do
			-- Is the tank alive (or died this turn)
			if _TankList[iter].hitPoints > 0 or (_TankList[iter].hitPoints == .5 and _TurnCount == _TankList[iter].turnKilled) then
				-- Yes
				-- Process tank moves
				-- Body Rotation
				if math.abs(_TankList[iter].playerCode.bodyRotationDelta) <= .1 then
				
					-- TODO 0 RAT Slow body rotation
					_TankList[iter].playerCode.bodyRotationDelta = _TankList[iter].playerCode.bodyRotationDelta * .5
				
					_TankList[iter].bodyRotation = _TankList[iter].bodyRotation + _TankList[iter].playerCode.bodyRotationDelta
					if _TankList[iter].bodyRotation < 0 then
						_TankList[iter].bodyRotation = _TankList[iter].bodyRotation + math.pi * 2
					elseif _TankList[iter].bodyRotation > math.pi * 2 then
						_TankList[iter].bodyRotation = _TankList[iter].bodyRotation - math.pi * 2
					end
				end
				-- Body Movement
				-- Did the tank move?
				_TankList[iter].playerCode.bodySpeed = math.floor(_TankList[iter].playerCode.bodySpeed)
				if _TankList[iter].playerCode.bodySpeed ~= 0 then
					-- Yes
					-- Find tanks close to this one
					local closeTankList = {}
					for closeIter = 1, #_TankList, 1 do
						-- Is this an enemy tank?
						if closeIter ~= iter then
							-- Is the enemy alive?
							if _TankList[closeIter].hitPoints > 0 then
								-- Is the enemy tank close
								if isTankClose(_TankList[closeIter].X, _TankList[closeIter].Y, _TankList[iter].X, _TankList[iter].Y, _TankList[iter].size * 2) then
									-- Yes, tank is close (within bullet "speed")
									-- table.insert(closeTankList, {closeIter, "Tank"})
									insertIfNotPresent(closeTankList, {closeIter, "Tank"})
								end
							end
						end
					end
					if _TankList[iter].playerCode.bodySpeed >= -3 and _TankList[iter].playerCode.bodySpeed <= 5 then
						local futureX = _TankList[iter].X
						local futureY = _TankList[iter].Y
						local canMove = true
						
						-- TODO 0 RAT Slow tank movement
						_TankList[iter].playerCode.bodySpeed = _TankList[iter].playerCode.bodySpeed * .5
						
						-- Is the tank moving forward?
						if _TankList[iter].playerCode.bodySpeed > 0 then
							-- Yes
							for pointIter = 1, _TankList[iter].playerCode.bodySpeed, 1 do
								futureX = futureX + (pointIter * math.sin(_TankList[iter].bodyRotation))
								futureY = futureY + (-pointIter * math.cos(_TankList[iter].bodyRotation))
								-- Is the move out of bounds?
								if futureX <= 10 or futureX >= (love.graphics.getWidth() - 10) or futureY <= 10 or futureY >= (love.graphics.getHeight() - 10) then
									-- Yes
									canMove = false
								else
									-- Would this hit another tank?
									for closeIter = 1, #closeTankList, 1 do
										objectX = _TankList[closeTankList[closeIter][1]].X
										objectY = _TankList[closeTankList[closeIter][1]].Y
										objectSize = _TankList[closeTankList[closeIter][1]].size
										if didCollideWithObject(futureX, futureY, objectX, objectY, objectSize) then
											canMove = false
										end
									end
								end
								if canMove then
									_TankList[iter].X = futureX
									_TankList[iter].Y = futureY
								else
									break
								end
							end
						else
							-- No, the tank is moving backward
--							print(_TankList[iter].name)
							-- TODO 0 bug tanks can't move backward
							for pointIter = 1, _TankList[iter].playerCode.bodySpeed, -1 do
								futureX = futureX + (pointIter * math.sin(_TankList[iter].bodyRotation))
								futureY = futureY + (-pointIter * math.cos(_TankList[iter].bodyRotation))
								-- Is the move out of bounds?
								if futureX <= 10 or futureX >= (love.graphics.getWidth() - 10) or futureY <= 10 or futureY >= (love.graphics.getHeight() - 10) then
									-- Yes
									canMove = false
								else
									-- Would this hit another tank?
									for closeIter = 1, #closeTankList, 1 do
										objectX = _TankList[closeTankList[closeIter][1]].X
										objectY = _TankList[closeTankList[closeIter][1]].Y
										objectSize = _TankList[closeTankList[closeIter][1]].size
										if didCollideWithObject(futureX, futureY, objectX, objectY, objectSize) then
											canMove = false
										end
									end
								end
								if canMove then
									_TankList[iter].X = futureX
									_TankList[iter].Y = futureY
								else
									break
								end
							end
						end
					end
				end
				-- Turret Movement
				if math.abs(_TankList[iter].playerCode.turretDelta) <= .1 then
				
					-- TODO 0 RAT Slow tank rotation
					_TankList[iter].playerCode.turretDelta =_TankList[iter].playerCode.turretDelta *.5
				
					_TankList[iter].turretRotation = _TankList[iter].turretRotation + _TankList[iter].playerCode.turretDelta
					if _TankList[iter].turretRotation < 0 then
						_TankList[iter].turretRotation = _TankList[iter].turretRotation + 2 * math.pi
					elseif _TankList[iter].turretRotation >= 2 * math.pi then
						_TankList[iter].turretRotation = _TankList[iter].turretRotation - 2 * math.pi
					end
				end
				-- Shoot
				if _TankList[iter]:fire() then
					table.insert(_BulletList, {_TankList[iter].X + (-5 * math.sin(_TankList[iter].bodyRotation)), _TankList[iter].Y + (5 * math.cos(_TankList[iter].bodyRotation)), _TankList[iter].turretRotation, iter})
				end
			end
		end
	end
end

function M:keypressed(key, isrepeat)
	if _StepMode then
		if key == "a" then
			_RunLoop = true
		elseif key == " " then
			_StepMode = false
			_RunLoop = true
		end
	else
		if key == " " then
			_RunLoop = false
			_StepMode = true
		end
	end
	-- Display the correct instruction text
	if _StepMode then
		_Arena:displayStepInstruction()
	else
		_Arena:displayFreezeInstruction()
	end
end

function M:keyreleased(key)
end

-- LOGIC
function checkListForValue(list, value)
	for iter = 1, #list, 1 do
		if list[iter] == value then
			return true
		end
	end
	return false
end

-- Return TRUE if the points x2 and y2 are within BULLET_SPEED distance to the points x and y.
function isObjectClose(x, y, x2, y2)
	if (math.sqrt(math.pow(x - x2, 2) + math.pow(y - y2, 2))) <= BULLET_SPEED then
		return true
	else
		return false
	end
end

function isTankClose(x, y, x2, y2, size)
	if (math.sqrt(math.pow(x - x2, 2) + math.pow(y - y2, 2))) <= size then
		return true
	else
		return false
	end
end

-- Return the distance between two points.
function distanceBetweenPoints(x, y, x2, y2)
	return math.sqrt(math.pow(x - x2, 2) + math.pow(y - y2, 2))
end

function didCollideWithObject(x, y, ox, oy, objectSize)
	if (math.sqrt(math.pow(x - ox, 2) + math.pow(y - oy, 2))) <= objectSize then
		return true
	else
		return false
	end
end

-- Add the "item" to the table if it does not already contain the item.
function insertIfNotPresent(list, item)
	if table.contains(list, item) == false then
		table.insert(list, item)
	end
end

function table.contains(table, element)
  for _, value in pairs(table) do
    if value == element then
      return true
    end
  end
  return false
end

function permute(tab, r, count)
	local r = r or #tab
	for i = 1, count or r do
		local j = math.random(i, r)
		tab[i], tab[j] = tab[j], tab[i]
	end
	return tab
end

function getRangeTable(start, stop) 
    local range = {}
    for i = start, stop, 20 do
        table.insert(range, i)
    end
    return range
end

-- In place sort of a table.
-- Usage: print(unpack(quicksort{5, 2, 7, 3, 4, 7, 1}))
-- NOTE: We have no idea what the start and endi arguments do.
function quicksort(t, start, endi)
  start, endi = start or 1, endi or #t
  --partition w.r.t. first element
  if(endi - start < 2) then return t end
  local pivot = start
  for i = start + 1, endi do
    if t[i] <= t[pivot] then
      local temp = t[pivot + 1]
      t[pivot + 1] = t[pivot]
      if(i == pivot + 1) then
        t[pivot] = temp
      else
        t[pivot] = t[i]
        t[i] = temp
      end
      pivot = pivot + 1
    end
  end
  t = quicksort(t, start, pivot - 1)
  return quicksort(t, pivot + 1, endi)
end

return M